/* ccgo: go/arg.cc
 *
 * Copyright 2004 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <arg.hh>
#include <iostream>
#include <string>
#include <sstream>

using namespace arg;
using namespace std;

bool V_int::set(const std::string & str)
{
	istringstream s(str);
	s >> v;
	if (s.bad() || ! s.eof()) return false;
	return true;
}

V_int::operator int()
{
	return v;
}

bool V_double::set(const std::string & str)
{
	istringstream s(str);
	s >> v;
	if (s.bad() || ! s.eof()) return false;
	return true;
}

V_double::operator double()
{
	return v;
}

bool V_string::set(const std::string & str)
{
	istringstream s(str);
	s >> v;
	if (s.bad() || ! s.eof()) return false;
	return true;
}

V_string::operator string()
{
	return v;
}

void Parser::add_opt(int key, string name, Value * store, bool optional)
{
	Opt o;
	o.k = key;
	o.n = name;
	o.s = store;
	o.o = optional;
	o.p = false;
	l.push_back(o);
}

bool Parser::parse(int argc, char * argv[])
{
	new_arg();
	for (vector<Opt>::iterator i = l.begin(); i != l.end(); i ++) i->p = false;
	for (int i = 0; i < argc; i ++) {
		string s = argv[i];
		if (s[0] != '-') { // non-option => argument
			add_arg(s);
			continue;
		}
		if (s[1] == '-') { // long options
			string n;
			string v;
			bool vp = false;
			int k = s.find('=');
			if (k != string::npos) { // value present
				n = s.substr(2, k - 2);
				v = s.substr(k + 1);
				vp = true;
			} else {
				n = s.substr(2);
			}
			vector<Opt>::iterator j;
			for (j = l.begin(); j != l.end(); j ++) if (j->n == n) { // find matching name
				j->p = true; // set present
				if (! j->s) {
					if (vp) { // unexpected value
						cerr << "unexpected value in " << s << "\n";
						return false;
					}

					break;
				}
				if (! (vp || j->o)) { // missing value
					cerr << "missing value in " << s << "\n";
					return false;
				}
				if (vp && ! j->s->set(v)) { // malformed value
					cerr << "malformed value in " << s << "\n";
					return false;
				}
				break;
			}
			if (j == l.end()) { // name not found!
				cerr << "unknown option \"" << s << "\"\n";
				return false;
			}
			continue;
		}
		// short options
		for (int k = 1; k < s.length(); k ++) {
			vector<Opt>::iterator j = l.begin();
			while (j != l.end()) {
				if (j->k == s[k]) break;
				j ++;
			}
			if (j == l.end()) { // key not found!
				cerr << "unknow key in " << s << "\n";
				return false;
			}
			j->p = true;
			if (! j->s) continue; // no value needed
			string v = s.substr(k + 1);
			if (v != "") { // string following valued option => set the value
				if (! j->s->set(v)) { // malformed value
					cerr << "malformed value in " << s << "\n";
					return false;
				}
				break;
			}
			// no string
			if (j->o) break; // optional, assume no value present
			// value required => should be the next argument
			i ++;
			if (i >= argc) { // missing required value
				cerr << "missing required value in " << s << "\n";
				return false;
			}
			v = argv[i];
			if (! j->s->set(v)) { // malformed value
				cerr << "malformed value in " << s << ' ' << v << "\n";
				return false;
			}
			break;
		}
	}
	return true;
}

bool Parser::check(int key)
{
	for (vector<Opt>::const_iterator i = l.begin(); i != l.end(); i ++) if (i->k == key) return i->p;
	return false;
}

const vector<string> & Parser::get_args() const
{
	return a;
}

void Parser::new_arg()
{
	a.clear();
}

void Parser::add_arg(const string & str)
{
	a.push_back(str);
}
